#!/usr/bin/env python3

"""A textual base app with common features like a logging window"""
# pylint: disable=duplicate-code
# pylint: disable=too-many-locals
# pylint: disable=too-many-arguments
# pylint: disable=too-few-public-methods
# pylint: disable=too-many-instance-attributes

import asyncio
import logging
from argparse import ArgumentParser
from collections.abc import MutableSequence, Sequence
from pathlib import Path

from rich.color import Color
from rich.console import Console, ConsoleOptions, RenderResult
from rich.logging import RichHandler
from rich.markup import escape as markup_escape
from rich.segment import Segment
from rich.style import Style
from rich.text import Text
from textual import on
from textual.app import App, ComposeResult
from textual.message import Message
from textual.renderables._blend_colors import blend_colors
from textual.renderables.sparkline import Sparkline as SparklineRenderable
from textual.scrollbar import ScrollTo
from textual.widgets import Label, RichLog

from .logging_helper import (
    LogLevelSpec,
    apply_common_logging_cli_args,
    callstack_filter,
    logger_name_filter,
    markup_escape_filter,
    set_log_levels,
    thread_id_filter,
)


def log() -> logging.Logger:
    """Returns the logger instance to use here"""
    return logging.getLogger("trickkiste.base_app")


class RichLogHandler(RichHandler):
    """Redirects rich.RichHanlder capabilities to a textual.RichLog"""

    def __init__(self, widget: RichLog, level: int = logging.INFO):
        super().__init__(show_path=False, markup=True, show_time=False, level=level)
        self.widget: RichLog = widget

    def emit(self, record: logging.LogRecord) -> None:
        record.args = record.args and tuple(
            markup_escape(arg) if isinstance(arg, str) else arg for arg in record.args
        )
        record.msg = markup_escape(record.msg)
        self.widget.write(
            self.render(
                record=record,
                message_renderable=self.render_message(record, self.format(record)),
                traceback=None,
            )
        )


class LockingRichLog(RichLog):
    """A RichLog which turns off autoscroll when scrolling manually"""

    @on(ScrollTo)
    def on_scroll_to(self, _event: Message) -> None:
        """Mandatory comment"""
        self.auto_scroll = self.is_vertical_scroll_end


class TuiBaseApp(App[None]):
    """A nice UI for Sauron stuff"""

    CSS_PATH = Path(__file__).parent / "base_tui_app.css"

    def __init__(
        self,
        logger_show_level: bool = True,
        logger_show_time: bool = True,
        logger_show_name: bool = True,
        logger_show_callstack: bool = False,
        logger_show_funcname: bool = False,
        logger_show_tid: bool = False,
    ) -> None:
        super().__init__()
        self._richlog = LockingRichLog(id="app_log")
        self._logger_show_level = logger_show_level
        self._logger_show_time = logger_show_time
        self._logger_show_name = logger_show_name
        self._logger_show_callstack = logger_show_callstack
        self._logger_show_funcname = logger_show_funcname
        self._logger_show_tid = logger_show_tid
        self._log_level: Sequence[LogLevelSpec] = (logging.INFO,)
        self._footer_label = Label(Text.from_markup("nonsense"), id="footer")

    def add_default_arguments(self, parser: ArgumentParser) -> None:
        """Adds arguments to @parser we need in every app"""
        apply_common_logging_cli_args(parser)

    def compose(self) -> ComposeResult:
        """Set up the UI"""
        yield self._richlog
        yield self._footer_label

    async def on_mount(self) -> None:
        """UI entry point"""
        logging.getLogger().handlers = [handler := RichLogHandler(self._richlog)]

        opt_time = "│ %(asctime)s " if self._logger_show_time else ""
        opt_name = "│ [grey53]%(name)-16s[/] " if self._logger_show_name else ""
        opt_funcname = "│ [grey53]%(funcName)-32s[/] " if self._logger_show_funcname else ""
        opt_callstack = "│ [grey53]%(callstack)-32s[/] " if self._logger_show_callstack else ""
        if self._logger_show_callstack:
            handler.addFilter(callstack_filter)
        opt_tid = "│ [grey53]%(posixTID)-8s[/] " if self._logger_show_tid else ""
        if self._logger_show_tid:
            handler.addFilter(thread_id_filter)
        handler.addFilter(markup_escape_filter)
        handler.addFilter(logger_name_filter)
        handler.setFormatter(
            logging.Formatter(
                f"{opt_time}{opt_tid}{opt_name}{opt_funcname}{opt_callstack}"
                "│ [bold white]%(message)s[/]",
                datefmt="%Y-%m-%d %H:%M:%S",
            )
        )

        self.set_log_levels(*self._log_level)

        await self.initialize()

    async def initialize(self) -> None:
        """Not implemented"""

    def update_status_bar(self, text: str) -> None:
        """Convenience wrapper - should go to TUIBaseApp"""
        self._footer_label.update(text)

    def execute(self) -> None:
        """Wrapper for async run and optional cleanup if provided"""
        asyncio.run(self.run_async())
        self.cleanup()

    def cleanup(self) -> None:
        """Not implemented"""

    def set_log_levels(
        self, *levels: LogLevelSpec, others_level: int | str = logging.WARNING
    ) -> None:
        """Sets the overall log level for internal log console"""
        set_log_levels(*levels, others_level=others_level)
        self._log_level = levels


class HeatBar(SparklineRenderable[float]):
    """SparklineRenderable with additional features"""

    BARS = "▁▂▃▄▅▆▇█"
    BARS_INVERTED = " 🮂🮃🮄🮅🮆█"

    def __init__(
        self,
        *,
        width: int | None = 10,
        min_color: int | str | Color = Color.from_rgb(0, 255, 0),
        max_color: int | str | Color = Color.from_rgb(255, 0, 0),
        min_bar_value: float | None = None,
        max_bar_value: float | None = None,
        min_color_value: float | None = None,
        max_color_value: float | None = None,
        inverted: bool = False,
    ) -> None:
        super().__init__(
            data=[],
            width=width,
            min_color=(
                Color.from_triplet(Color.from_ansi(min_color).get_truecolor())
                if isinstance(min_color, int)
                else (
                    Color.from_triplet(Color.parse(min_color).get_truecolor())
                    if isinstance(min_color, str)
                    else min_color
                )
            ),
            max_color=(
                Color.from_triplet(Color.from_ansi(max_color).get_truecolor())
                if isinstance(max_color, int)
                else (
                    Color.from_triplet(Color.parse(max_color).get_truecolor())
                    if isinstance(max_color, str)
                    else max_color
                )
            ),
        )
        self.data: MutableSequence[float] = []
        self.bars = self.BARS_INVERTED if inverted else self.BARS
        self.inverted = inverted
        self.min_bar_value = min_bar_value
        self.max_bar_value = max_bar_value
        self.min_color_value = min_color_value
        self.max_color_value = max_color_value

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        width = self.width or options.max_width
        len_data = len(self.data)
        if len_data == 0:
            yield Segment("▁" * width, self.min_color)
            return
        if len_data == 1:
            yield Segment("█" * width, self.max_color)
            return

        minimum = min(self.data) if self.min_bar_value is None else self.min_bar_value
        maximum = max(self.data) if self.max_bar_value is None else self.max_bar_value

        extent = maximum - minimum or 1

        buckets = tuple(self._buckets(list(self.data), num_buckets=width))

        bucket_index = 0.0
        bars_rendered = 0
        step = len(buckets) / width
        summary_function = self.summary_function
        min_color, max_color = self.min_color.color, self.max_color.color
        assert min_color is not None
        assert max_color is not None
        while bars_rendered < width:
            partition = buckets[int(bucket_index)]
            partition_summary = summary_function(partition)
            height_ratio = (partition_summary - minimum) / extent
            bar_index = int(height_ratio * (len(self.bars) - 1))
            bar_color = blend_colors(min_color, max_color, height_ratio)
            bars_rendered += 1
            bucket_index += step
            yield Segment(
                self.bars[min(bar_index, len(self.bars) - 1)], Style.from_color(bar_color)
            )



#!/usr/bin/env python3

"""Example for fancy logging"""

import logging
from argparse import ArgumentParser

from trickkiste.logging_helper import (
    apply_common_logging_cli_args,
    set_log_levels,
    setup_logging,
)


def log() -> logging.Logger:
    """Returns the logger instance to use here"""
    return logging.getLogger("trickkiste.fancylogging")


def main() -> None:
    """Runs this"""
    parser = ArgumentParser(__doc__)
    apply_common_logging_cli_args(parser)
    args = parser.parse_args()

    setup_logging(log(), level=args.log_level)

    set_log_levels("DEBUG")
    set_log_levels(logging.DEBUG)
    set_log_levels((log(), "DEBUG"))
    set_log_levels(("trickkiste", "INFO"), (log(), "DEBUG"))

    logging.getLogger().debug("foo")
    logging.getLogger().info("foo")
    logging.getLogger().warning("foo")
    logging.getLogger().error("foo")

    logging.getLogger("trickkiste").debug("debug")
    logging.getLogger("trickkiste").info("info")
    logging.getLogger("trickkiste").warning("warning")
    logging.getLogger("trickkiste").error("error")

    log().debug("debug")
    log().info("info")
    log().warning("warning")
    log().error("error")

    logging.getLogger("other.module").debug("only shown with level=ALL_DEBUG")
    logging.getLogger("other.module").info("only shown with level=ALL_DEBUG")
    logging.getLogger("other.module").warning("this is visible")
    logging.getLogger("other.module").error("this is also visible")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""Common stuff shared among modules"""

import logging
import os
import sys
import threading
import traceback
from argparse import ArgumentParser
from collections.abc import Iterable

from rich.console import Console
from rich.logging import RichHandler
from rich.markup import escape as markup_escape

LOG_LEVELS = ("ALL_DEBUG", "DEBUG", "INFO", "WARN", "ERROR", "CRITICAL")
LogLevelSpec = int | str | tuple[str | logging.Logger, int | str]


def apply_common_logging_cli_args(parser: ArgumentParser) -> None:
    """Decorates given @parser with arguments for logging"""
    parser.add_argument(
        "--log-level",
        "-l",
        choices=LOG_LEVELS,
        help="Sets the logging level - ALL_DEBUG sets all other loggers to DEBUG, too",
        type=str.upper,
        default="INFO",
    )


def stack_str(depth: int = 0) -> str:
    """Returns a short local function call stack"""

    def stack_fns() -> Iterable[str]:
        stack = list(
            reversed(
                traceback.extract_stack(sys._getframe(depth))  # pylint: disable=protected-access
            )
        )

        for site in stack:
            if site.filename != stack[0].filename or site.name == "<module>":
                break
            yield site.name

    return ">".join(reversed(list(stack_fns())))


def markup_escape_filter(record: logging.LogRecord) -> bool:
    """Escapes log record contents in order to avoid unintended formatting"""
    record.args = record.args and tuple(
        markup_escape(arg) if isinstance(arg, str) else arg for arg in record.args
    )
    record.msg = markup_escape(record.msg)
    return True


def thread_id_filter(record: logging.LogRecord) -> bool:
    """Inject thread_id in log records"""
    record.posixTID = threading.get_native_id()
    return True


def callstack_filter(record: logging.LogRecord) -> bool:
    """Inject short function call stack in log records"""
    record.callstack = stack_str(5)
    return True


def logger_name_filter(record: logging.LogRecord) -> bool:
    """Inject thread_id to log records"""
    record.name = record.name[11:] if record.name.startswith("trickkiste.") else record.name
    return True


def setup_logging(  # pylint: disable=too-many-arguments
    logger: logging.Logger | str,
    level: str | int = "INFO",
    show_level: bool = True,
    show_time: bool = True,
    show_name: bool = True,
    show_callstack: bool = False,
    show_funcname: bool = True,
    show_tid: bool = False,
) -> None:
    """Make logging fun"""
    if not logging.getLogger().hasHandlers():
        handler = RichHandler(
            show_level=show_level,
            show_time=False,
            omit_repeated_times=True,
            show_path=False,
            markup=True,
            console=Console(
                stderr=True, color_system="standard" if os.environ.get("FORCE_COLOR") else "auto"
            ),
        )
        logging.getLogger().addHandler(handler)

        opt_time = "│ %(asctime)s " if show_time else ""
        opt_name = "│ [grey53]%(name)-16s[/] " if show_name else ""
        opt_funcname = "│ [grey53]%(funcName)-32s[/] " if show_funcname else ""
        opt_callstack = "│ [grey53]%(callstack)-32s[/] " if show_callstack else ""
        if show_callstack:
            handler.addFilter(callstack_filter)
        opt_tid = "│ [grey53]%(posixTID)-8s[/] " if show_tid else ""
        if show_tid:
            handler.addFilter(thread_id_filter)
        handler.addFilter(markup_escape_filter)
        handler.addFilter(logger_name_filter)

        handler.setFormatter(
            logging.Formatter(
                f"{opt_time}{opt_tid}{opt_name}{opt_funcname}{opt_callstack}"
                "│ [bold white]%(message)s[/]",
                datefmt="%Y-%m-%d %H:%M:%S",
            )
        )

    set_log_levels((logger, level))

    # for lev in LOG_LEVELS:
    #    logging.addLevelName(getattr(logging, lev), f"{lev[0] * 2}")

    logging.getLogger("urllib3.connectionpool").setLevel(logging.INFO)


def set_log_levels(*levels: LogLevelSpec, others_level: int | str = logging.WARNING) -> None:
    """Sets the overall log level for internal log console"""

    def level_of(level: str | int) -> int:
        return int(logging.getLevelName(level.split("_")[-1])) if isinstance(level, str) else level

    named_levels: dict[str | None, int] = {
        **{
            None: logging.DEBUG if "ALL_DEBUG" in levels else level_of(others_level),
            "trickkiste": logging.INFO,
        },
        **dict(
            (
                ("trickkiste", level_of(level_spec))
                if isinstance(level_spec, (int, str))
                else (
                    (
                        level_spec[0].name
                        if isinstance(level_spec[0], logging.Logger)
                        else level_spec[0]
                    ),
                    level_of(level_spec[1]),
                )
            )
            for level_spec in levels
        ),
    }

    for handler in logging.getLogger().handlers:
        handler.setLevel(min(named_levels.values()))

    for name, level in named_levels.items():
        logging.getLogger(name).setLevel(level)



#!/usr/bin/env python

"""Runs a process but suppress output if it returns successfully before a given amount of time"""

import signal
import sys
from asyncio import Queue, StreamReader
from asyncio import TimeoutError as AsyncTimeoutError
from asyncio import create_subprocess_exec, gather, run, wait_for
from asyncio.subprocess import PIPE, Process
from contextlib import suppress
from typing import Optional, Sequence, TextIO

LineQueue = Queue[Optional[tuple[TextIO, bytes]]]


async def print_after(
    timeout: float,
    abort: Queue[bool],
    buffer: LineQueue,
) -> None:
    """Wait for a given time or until aborted - print buffer contents if appropriate"""
    with suppress(AsyncTimeoutError):
        if await wait_for(abort.get(), timeout):
            return
    while elem := await buffer.get():
        out_file, line = elem
        out_file.write(line.decode(errors="replace"))


async def buffer_stream(stream: StreamReader, buffer: LineQueue, out_file: TextIO) -> None:
    """Records a given stream to a buffer line by line along with the source"""
    while line := await stream.readline():
        await buffer.put((out_file, line))
    await buffer.put(None)


async def wait_and_notify(process: Process, abort: Queue[bool]) -> None:
    """Just waits for @process to finish and notify the result"""
    await process.wait()
    await abort.put(process.returncode == 0)


async def run_quiet_and_verbose(timeout: float, cmd: Sequence[str]) -> None:
    """Run a command and start printing it's output only after a given timeout"""
    buffer: LineQueue = Queue()
    abort: Queue[bool] = Queue()

    process = await create_subprocess_exec(*cmd, stdout=PIPE, stderr=PIPE)

    assert process.stdout
    assert process.stderr

    signal.signal(signal.SIGINT, lambda _sig, _frame: 0)

    await gather(
        print_after(float(timeout), abort, buffer),
        buffer_stream(process.stdout, buffer, sys.stdout),
        buffer_stream(process.stderr, buffer, sys.stderr),
        wait_and_notify(process, abort),
    )
    raise SystemExit(process.returncode)


def main() -> None:
    """Just the entrypoint for run_quiet_and_verbose()"""
    timeout, *cmd = sys.argv[1:]
    run(run_quiet_and_verbose(float(timeout), cmd))


if __name__ == "__main__":
    main()
